# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2020, Gaute Hope, MET Norway

from opendrift.readers.basereader import BaseReader, ContinuousReader

import warnings
import pyproj
import numpy as np
import shapely.vectorized
import shapely.prepared
from shapely.geometry import box
import shapely.wkb as wkb
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import logging

logger = logging.getLogger(__name__)

# TODO: replace this with weakref + thread-safety
__roaring_mask__ = None
__polys__ = None

def get_mask():
    """
    Returns an instance of the landmask type and landmask. The mask data is
    usually shared between threads.
    """
    global __roaring_mask__

    if __roaring_mask__ is None:
        from roaring_landmask import RoaringLandmask
        __roaring_mask__ = RoaringLandmask.new()

    return __roaring_mask__

class LandmaskFeature(cfeature.GSHHSFeature):
    def __init__(self, scale='auto', globe=None, **kwargs):
        super().__init__(scale, **kwargs)

        if globe is not None:
            self._crs = ccrs.PlateCarree(globe=globe)

    def geometries(self):
        self.intersecting_geometries(extent=None)

    def intersecting_geometries(self, extent):
        global __polys__

        if self._scale == 'auto':
            scale = self._scale_from_extent(extent)
        else:
            scale = self._scale[0]

        # # If scale is full use the geometries from roaring landmask, otherwise
        # # fall back to Cartopy provider.
        # if scale == 'f':
        #     logger.debug(f"Adding full GSHHG shapes from roaring-landmask, extent: {extent}..")
        #     if __polys__ is None:
        #         from roaring_landmask import Gshhg
        #         polys = Gshhg.wkb()
        #         __polys__ = wkb.loads(polys)

        #     if extent is not None:
        #         extent = box(extent[0], extent[2], extent[1], extent[3])
        #         extent = shapely.prepared.prep(extent)
        #         return (p for p in __polys__.geoms if extent.intersects(p))
        #     else:
        #         return __polys__.geoms
        # else:
        logger.debug(f"Adding GSHHG shapes from cartopy, scale: {scale}, extent: {extent}..")
        return super().intersecting_geometries(extent)

def plot_land(ax, lonmin, latmin, lonmax, latmax, fast, ocean_color = 'white', land_color = cfeature.COLORS['land'], lscale = 'auto', globe=None):
    """
    Plot the landmask or the shapes from GSHHG.
    """
    def show_landmask_roaring(roaring):
        maxn = 512.
        dx = (lonmax - lonmin) / maxn
        dy = (latmax - latmin) / maxn
        dx = max(roaring.dx, dx)
        dy = max(roaring.dy, dy)

        x = np.arange(lonmin, lonmax, dx)
        y = np.arange(latmin, latmax, dy)

        yy, xx = np.meshgrid(y, x)
        img = roaring.mask.contains_many_par(xx.ravel(), yy.ravel()).reshape(yy.shape).T

        from matplotlib import colors
        cmap = colors.ListedColormap([ocean_color, land_color])
        ax.imshow(img, origin = 'lower', extent=[lonmin, lonmax, latmin, latmax],
                  zorder=0,
                    transform=ccrs.PlateCarree(globe=globe), cmap=cmap)
    if fast:
        show_landmask_roaring(get_mask())
    else:
        land = LandmaskFeature(scale=lscale, facecolor=land_color, globe=globe)

        ax.add_feature(land, zorder=2,
                       facecolor=land_color,
                       edgecolor='black')



class Reader(BaseReader, ContinuousReader):
    """
    The global landmask reader is based on the coastline data from
    GSHHG (https://www.ngdc.noaa.gov/mgg/shorelines/) optimized for
    checking against landmasks.
    """
    name = 'global_landmask'
    variables = ['land_binary_mask']
    proj4 = None
    crs = None

    def __init__(self):
        self.proj4 = '+proj=lonlat +ellps=WGS84'
        self.crs = pyproj.CRS(self.proj4)

        super(Reader, self).__init__()

        # Depth
        self.z = None

        self.xmin, self.ymin = -180, -90
        self.xmax, self.ymax = 180, 90

        # setup landmask
        self.mask = get_mask()

    def __on_land__(self, x, y):
        x = self.modulate_longitude(x)
        x = x.astype(np.float64)
        y = y.astype(np.float64)
        return self.mask.contains_many_par(x, y)

    def get_variables(self,
                      requestedVariables,
                      time=None,
                      x=None,
                      y=None,
                      z=None):
        """
        Get binary mask of whether elements are on land or not.

        Args:
            x (deg[]): longitude (decimal degrees)
            y (deg[]): latitude (decimal degrees)
            ...

        x, y is given in reader local projection.

        Returns:
            Binary mask of point x, y on land.

        """

        self.check_arguments(requestedVariables, time, x, y, z)
        return {'land_binary_mask': self.__on_land__(x, y)}



#! /usr/bin/env python
#
# Author: Gaute Hope (gauteh@met.no) / 2022
#
# Converts oils from old CSV format to JSON format using `adios_db` (make sure it is installed from adios_oil_database).
#
# Usage:
#
#   csv_to_json.py CSVFILE outdir/

import sys
import logging
import coloredlogs
from pathlib import Path

def main():
    from adios_db.data_sources.noaa_fm import OilLibraryCsvFile, OilLibraryRecordParser, OilLibraryAttributeMapper
    from adios_db.models.oil.validation.validate import validate_json
    from adios_db.models.oil.completeness import set_completeness

    logger = logging.getLogger(__name__)
    coloredlogs.install('debug')

    CSV = sys.argv[1]
    OUT = sys.argv[2]

    logger.info(f"Reading input file: {CSV}")
    logger.info(f"Saving JSON files to: {OUT}")

    config = None
    reader = OilLibraryCsvFile
    parser = OilLibraryRecordParser
    mapper = OilLibraryAttributeMapper

    for record in reader(CSV).get_records():
        oil_map = mapper(parser(*record))
        oil_json = oil_map.py_json()
        oil = validate_json(oil_json)
        set_completeness(oil)

        outf = Path(OUT).joinpath(f"{oil.oil_id}.json")
        logger.info(f"Writing {outf}..")
        with open(outf, 'w') as fd:
            oil.to_file(outf)

if __name__ == '__main__':
    main()




# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2022, Knut-Frode Dagestad, MET Norway

"""
OpenHNS is a 3D HNS drift module bundled within the OpenDrift framework.
"""

from io import open
import numpy as np
from datetime import datetime
import pyproj
import matplotlib.pyplot as plt
import logging

logger = logging.getLogger(__name__)

from opendrift.models.oceandrift import OceanDrift, Lagrangian3DArray


# Defining the oil element properties
class HNS(Lagrangian3DArray):
    """Extending LagrangianArray with variables relevant for HNS particles."""

    variables = Lagrangian3DArray.add_variables([
        (   'mass', {
                'dtype': np.float32,
                'units': 'kg',
                'seed': False,
                'default': 1
        }),
        (
            'viscosity',
            {
                'dtype': np.float32,
                'units': 'N s/m2 (Pa s)',
                'seed': False,  # Taken from NOAA database
                'default': 0.005
            }),
        (
            'density',
            {
                'dtype': np.float32,
                'units': 'kg/m^3',
                'seed': False,  # Taken from NOAA database
                'default': 880
            }),
        (
            'wind_drift_factor',
            {
                'dtype':
                np.float32,  # TODO: inherit from
                'units':
                '%',  # OceanDrift
                'description':
                'Elements at the ocean surface are moved by '
                'this fraction of the wind vector, in addition to '
                'currents and Stokes drift',
                'default':
                0.03
            }),
        (
           'diameter',
            {
                'dtype': np.float32,  # Particle diameter
                'units': 'm',
                'seed': False,
                'default': 0.
            })
    ])


class OpenHNS(OceanDrift):
    """Open source HNS drift model based on the OpenDrift framework.

        Developed at MET Norway based on parameterisations
        found in open/published litterature.

        Under construction.
    """

    ElementType = HNS

    required_variables = {
        'x_sea_water_velocity': {
            'fallback': None
        },
        'y_sea_water_velocity': {
            'fallback': None
        },
        'x_wind': {
            'fallback': None
        },
        'y_wind': {
            'fallback': None
        },
        'upward_sea_water_velocity': {
            'fallback': 0,
            'important': False
        },
        'sea_surface_wave_significant_height': {
            'fallback': 0,
            'important': False
        },
        'sea_surface_wave_stokes_drift_x_velocity': {
            'fallback': 0,
            'important': False
        },
        'sea_surface_wave_stokes_drift_y_velocity': {
            'fallback': 0,
            'important': False
        },
        'sea_surface_wave_period_at_variance_spectral_density_maximum': {
            'fallback': 0,
            'important': False
        },
        'sea_surface_wave_mean_period_from_variance_spectral_density_second_frequency_moment':
        {
            'fallback': 0,
            'important': False
        },
        'sea_ice_area_fraction': {
            'fallback': 0,
            'important': False
        },
        'sea_ice_x_velocity': {
            'fallback': 0,
            'important': False
        },
        'sea_ice_y_velocity': {
            'fallback': 0,
            'important': False
        },
        'sea_water_temperature': {
            'fallback': 10,
            'profiles': True
        },
        'sea_water_salinity': {
            'fallback': 34,
            'profiles': True
        },
        'sea_floor_depth_below_sea_level': {
            'fallback': 10000
        },
        'ocean_vertical_diffusivity': {
            'fallback': 0.02,
            'important': False,
            'profiles': True
        },
        'land_binary_mask': {
            'fallback': None
        },
        'ocean_mixed_layer_thickness': {
            'fallback': 50,
            'important': False
        },
    }

    # The depth range (in m) which profiles shall cover
    required_profiles_z_range = [-20, 0]

    max_speed = 1.3  # m/s

    hnstypes = ['chem1', 'chem2']

    # Default colors for plotting
    status_colors = {
        'initial': 'green',
        'active': 'blue',
        'missing_data': 'gray',
        'stranded': 'red',
        'evaporated': 'yellow',
        'dispersed': 'magenta'
    }


    def __init__(self, *args, **kwargs):

        # Calling general constructor of parent class
        super(OpenHNS, self).__init__(*args, **kwargs)

        self._add_config({
            'seed:evaporation_rate': {
                'type': 'float',
                'default': .99,
                'min': 0,
                'max': 1e10,
                'units': 's-1',
                'description':
                'The evaporation rate', 
                'level': self.CONFIG_LEVEL_ESSENTIAL,
                'description': 'Evaporation rate'
            },
           'seed:entrainment_rate': {
                'type': 'float',
                'default': 1,
                'min': 0,
                'max': 1e10,
                'units': '1',
                'description':
                'The evaporation rate', 
                'level': self.CONFIG_LEVEL_ESSENTIAL,
                'description': 'Entrainment rate'
            },
           'seed:hns_type': {
                'type':
                'enum',
                'enum':
                self.hnstypes,
                'default':
                self.hnstypes[0],
                'level':
                self.CONFIG_LEVEL_ESSENTIAL,
                'description':
                'HNS type to be used for the simulation'
            },
        })

        self._set_config_default('drift:vertical_advection', False)
        self._set_config_default('drift:vertical_mixing', True)
        self._set_config_default('drift:current_uncertainty', 0.05)
        self._set_config_default('drift:wind_uncertainty', 0.5)

    def evaporation(self):
        print('Evaporating!')
        surface = np.where(self.elements.z == 0)[0]
        random_number = np.random.uniform(0, 1, len(surface))
        evaporated = np.where(random_number > self.get_config('seed:evaporation_rate'))[0]
        logger.debug('Evaporating %i of %i elements at ocean surface' % (len(evaporated), len(surface)))
        self.elements.wind_drift_factor[evaporated] = 1  # Shall follow wind 100%
        self.elements.z[evaporated] = 10  # Moving evaporated elements to 10m height


    def update(self):

        self.evaporation()
        #self.advect_ocean_current()
        self.advect_wind()
